/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"

/* #define PROFILING_CONTROL */

/* local */
static Scheme_Object *debug_source(int c, Scheme_Object *p[]);
#ifdef PROFILING_CONTROL
static Scheme_Object *profile_switch(int c, Scheme_Object *p[]);
#endif

void
scheme_init_debug (Scheme_Env *env)
{
  scheme_add_global_constant("debug-source", scheme_make_prim(debug_source), env);

#ifdef PROFILING_CONTROL
  scheme_add_global_constant("profile", scheme_make_prim(profile_switch), env);
#endif
}

Scheme_Debugging_Info *scheme_alloc_debug(Scheme_Object *src)
{
  Scheme_Debugging_Info *debug;

  debug = (Scheme_Debugging_Info*)scheme_malloc(sizeof(Scheme_Debugging_Info));
  
  debug->src = src;

  return debug;
}

Scheme_Object *scheme_debug_source(Scheme_Object *handle)
{
  Scheme_Debugging_Info *d;

  SCHEME_ASSERT(SCHEME_DEBUGP(handle), 
		"debug-source: expected a debug handle");

  d = SCHEME_DEBUG(handle);
  if (!d)
    return scheme_void;
  else
    return d->src;
}

static Scheme_Object *debug_source(int c, Scheme_Object *p[])
{
  SCHEME_ASSERT(c == 1, "debug-source: expected one argument");
  return scheme_debug_source(p[0]);
}

#ifdef PROFILING_CONTROL
static Scheme_Object *profile_switch(int c, Scheme_Object *p[])
{
  if (c != 1)
    scheme_wrong_count("profile", 1, 1, c, p);

  moncontrol(SCHEME_TRUEP(p[0]));

  return scheme_void;
}
#endif
