/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"

struct Scheme_Promise
{
  Scheme_Type type;
  int forced;
  Scheme_Object *val;
  Scheme_Env *env;
};
typedef struct Scheme_Promise Scheme_Promise;

/* globals */

/* locals */
static Scheme_Object *force (int argc, Scheme_Object *argv[]);

void
scheme_init_promise (Scheme_Env *env)
{
  scheme_add_global_constant ("force", 
			      scheme_make_prim_w_arity(force, 1, 1), 
			      env);
}

Scheme_Object *
scheme_make_promise_compiled(Scheme_Object *expr, Scheme_Env *env)
{
  Scheme_Promise *promise;

  promise = (Scheme_Promise *)scheme_malloc(sizeof(Scheme_Promise));
  promise->type = scheme_promise_type;
  promise->forced = 0;
  promise->val = expr;
  promise->env = env;

  return (Scheme_Object *)promise;
}

Scheme_Object *
scheme_make_promise(Scheme_Object *expr, Scheme_Env *env)
{
  return scheme_make_promise_compiled(scheme_compile(expr, env), env);
}

static Scheme_Object *
force (int argc, Scheme_Object *argv[])
{
  Scheme_Promise *promise;
  Scheme_Object *v;

  if (argc != 1)
    scheme_wrong_count("force", 1, 1, argc, argv);
  if (!SCHEME_PROMP(argv[0]))
    scheme_wrong_type("force", "promise", 0, argc, argv);

  promise = (Scheme_Promise *)argv[0];
  if (!promise->forced) {
    v = _scheme_eval_compiled(promise->val, promise->env);
    if (!promise->forced) {
      promise->val = v;
      promise->forced = 1;
    }
  }

  return promise->val;
}
