/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"
#include <string.h>
#include <memory.h>
#include "../gc/gc.h"

#ifdef NO_GC
#define MALLOC malloc
#else
#define MALLOC GC_malloc
#endif

#if 0
Scheme_Object *
scheme_alloc_object (void)
{
  Scheme_Object *object;

  object = (Scheme_Object *) scheme_malloc (sizeof (Scheme_Object));

  return (object);
}
#endif

#ifdef NO_GC
void *
scheme_malloc (size_t size)
{
  void *space;

  space = MALLOC (size);
  SCHEME_ASSERT ((space != 0), "memory allocation failure");
  return (space);
}
#endif

void *
scheme_calloc (size_t num, size_t size)
{
  void *space;
  
  space = MALLOC (num*size);
#ifdef NO_GC
  memset (space, 0, (num*size));
#endif
  SCHEME_ASSERT ((space != 0), "memory allocation failure");
  return (space);
}

char *
scheme_strdup(const char *str)
{
  char *naya;
  long len;

  len = strlen (str) + 1;
  naya = (char *)scheme_malloc_atomic (len * sizeof (char));
  memcpy (naya, str, len);

  return naya;
}

char *
scheme_strdup_eternal(const char *str)
{
  char *naya;
  long len;

  len = strlen (str) + 1;
  naya = (char *)scheme_malloc_eternal(len * sizeof (char));
  memcpy (naya, str, len);

  return naya;
}

void scheme_end_stubborn_change(void *p)
{
  GC_end_stubborn_change(p);
}

void *scheme_malloc_eternal(size_t n)
{
  void *s;

  s = malloc(n);
  memset(s, 0, n);

  return s;
}

#ifdef DOS_MEMORY

int scheme_same_pointer(void *a, void *b)
{
  long as, ao, bs, bo, areal, breal;
  
  as = FP_SEG(a);
  ao = FP_OFF(a);
  bs = FP_SEG(b);
  bo = FP_SEG(b);

  areal = (as << 4) + ao;
  breal = (bs << 4) + bo;

  return areal == breal;
}

int scheme_diff_pointer(void *a, void *b)
{
  return !scheme_same_pointer(a, b);
}

#endif
