/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "../sconfig.h"
#include "scheme.h"

typedef struct {
  Scheme_Bucket bucket;
  short const_val, keyword;
} Scheme_Bucket_With_Const_Flag;


typedef struct Scheme_Closure_Compilation_Data
{
  Scheme_Type type;
  /* Scheme_Object *src; */
  char must_allocate, has_rest;
  short num_params; /* includes collecting arg if has_rest */
  short extra_frame_slots;
  /* Scheme_Object *params; */
  Scheme_Object *code;
} Scheme_Closure_Compilation_Data;

typedef struct Scheme_Structure
{
  Scheme_Type type;
  char *type_name;
  int num_slots;
  Scheme_Object **slots;
} Scheme_Structure;

typedef struct Scheme_Start_Env
{
  struct Scheme_Basic_Env basic;
  Scheme_Hash_Table *globals;
} Scheme_Start_Env;

extern int scheme_in_macro_application;

extern int scheme_starting_up;

/* ininitializarion */
void scheme_init_stack_check(void);
Scheme_Process *scheme_make_process();
void scheme_top_level_env(Scheme_Process *process);
void scheme_init_process_error(Scheme_Process *process);
void scheme_init_type_symbol_table(void);
void scheme_init_symbol_table (void);
void scheme_init_symbol_type (Scheme_Env *env);
void scheme_init_type (Scheme_Env *env);
void scheme_init_list (Scheme_Env *env);
void scheme_init_port (Scheme_Env *env);
void scheme_init_proc (Scheme_Env *env);
void scheme_init_vector (Scheme_Env *env);
void scheme_init_string (Scheme_Env *env);
void scheme_init_number (Scheme_Env *env);
void scheme_init_eval (Scheme_Env *env);
void scheme_init_promise (Scheme_Env *env);
void scheme_init_struct (Scheme_Env *env);
void scheme_init_fun(Scheme_Env *env);
void scheme_init_symbol(Scheme_Env *env);
void scheme_init_type_symbol(Scheme_Env *env);
void scheme_init_nonvalue(Scheme_Env *env);
void scheme_init_char(Scheme_Env *env);
void scheme_init_bool(Scheme_Env *env);
void scheme_init_syntax(Scheme_Env *env);
void scheme_init_error(Scheme_Env *env);
void scheme_init_debug(Scheme_Env *env);
void scheme_init_object(Scheme_Env *env);
void scheme_init_process(Scheme_Env *env);
void scheme_init_trigger(Scheme_Env *env);
void scheme_init_dynamic_extension(Scheme_Env *env);
extern void scheme_regexp_initialize(Scheme_Env *env);

extern Scheme_Type_Reader *scheme_type_readers;
extern Scheme_Type_Writer *scheme_type_writers;

Scheme_Object *scheme_find_executer_name(Scheme_Syntax_Compilation *f);
Scheme_Syntax_Compilation *scheme_find_executer(Scheme_Object *sym);

extern Scheme_Object *scheme_void_func;

extern short scheme_can_read_compiled;

extern int scheme_fuel_counter;

Scheme_Object *scheme_read_number(const char *str, int is_float, 
				  int is_not_float,
				  int radix, int radix_set, int complain);

typedef long bignum;
typedef struct {
  Scheme_Object o;
  bignum v[1];
} Small_Bignum;
  
Scheme_Object *scheme_make_small_bignum(int v, Small_Bignum *s);
Scheme_Object *scheme_double_to_string(double d);

#define _scheme_apply(r,n,rs) scheme_do_eval(r,n,(Scheme_Env *)rs,1)
#define _scheme_apply_multi(r,n,rs) scheme_do_eval(r,n,(Scheme_Env *)rs,-1)
#define _scheme_tail_apply scheme_tail_apply
Scheme_Object *_scheme_apply_to_list (Scheme_Object *rator, Scheme_Object *rands);
Scheme_Object *_scheme_tail_apply_to_list (Scheme_Object *rator, Scheme_Object *rands);

#define _scheme_eval_compiled(obj, env) scheme_do_eval(obj, -1, env, 1)
#define _scheme_tail_eval scheme_tail_eval

void scheme_insure_stack_start(Scheme_Process *p, void *d);

Scheme_Object *scheme_top_level_eval(Scheme_Object *, int, Scheme_Object **);

typedef struct Scheme_Env_Stack_State {
  int stack_pos;
  struct Scheme_Env_Stack *env_stack;
} Scheme_Env_Stack_State;

typedef struct Scheme_Dynamic_Wind {
  void *data;
  void (*pre)(void *);
  void (*post)(void *);
  jmp_buf saveerr;
  struct Scheme_Env_Stack_State envss;
  struct Scheme_Cont *cont;
  struct Scheme_Dynamic_Wind *prev, *next;
} Scheme_Dynamic_Wind;



typedef struct Scheme_Cont {
  Scheme_Type type;
  Scheme_Object *value;
  Scheme_Jumpup_Buf buf;
  long *ok;
  Scheme_Dynamic_Wind *dw, *common;
  Scheme_Process *home;
  int orig_overflow;
  struct Scheme_Env_Stack_State envss;
  struct Scheme_Cont *prev;
  struct Scheme_Quick_Env *saved_env_stack;
  long saved_stack_size;
  struct Scheme_Overflow *save_overflow;
} Scheme_Cont;

typedef struct Scheme_Escaping_Cont {
  Scheme_Type type;
  Scheme_Process *home;
  Scheme_Object *val;
  Scheme_Object *f;
  int orig_overflow;
} Scheme_Escaping_Cont;

#define SCHEME_CONT_HOME(obj)  (((Scheme_Escaping_Cont *)(obj))->home)
#define SCHEME_CONT_VAL(obj)  (((Scheme_Escaping_Cont *)(obj))->val)

#define scheme_force_value(v) (((v == scheme_tail_call_waiting) || (v == scheme_eval_waiting)) ? scheme_value(v) : 0)

Scheme_Object *scheme_do_eval (Scheme_Object *obj, int _num_rands, Scheme_Env *env, int val);

Scheme_Object *scheme_value(Scheme_Object *);

#define _scheme_do_eval(obj, env, v) \
  ((SCHEME_INTP(obj) || !SCHEME_STRTAG_VAL(_SCHEME_TYPE(obj))) \
   ? obj : scheme_do_eval(obj, -1, env, v))
#define q_scheme_eval_compiled(obj, env) _scheme_do_eval(obj, env, 1)
#define q_scheme_tail_eval(obj, env) scheme_tail_eval(obj, env)

Scheme_Object *scheme_apply_struct_proc (Scheme_Object *rator, int c, Scheme_Object **rands);

/* tail calls (internal): */
extern Scheme_Object *scheme_eval_wait_expr;
extern Scheme_Env *scheme_eval_wait_env;
extern Scheme_Object *scheme_tail_rator;
extern int scheme_tail_num_rands;
extern Scheme_Object **scheme_tail_rands;

typedef struct Scheme_Trigger_Callback {
  int start;
  Scheme_Trigger *trigger;
  Scheme_Object *callback;
  struct Scheme_Trigger_Callback *next;
} Scheme_Trigger_Callback;

typedef struct Scheme_Trigger_From_Port {
  Scheme_Trigger *trigger;
  Scheme_Object *port;
  struct Scheme_Trigger_From_Port *next;
} Scheme_Trigger_From_Port;

typedef struct Scheme_Static_Dist {
  Scheme_Type type;
  short frame, position;
} Scheme_Static_Dist;

typedef struct Scheme_Let_Value {
  Scheme_Type type;
  int position;
  Scheme_Object *value;
  Scheme_Object *body;
} Scheme_Let_Value;

typedef struct Scheme_Let_Void {
  Scheme_Type type;
  short start, end;
  Scheme_Object *body;
} Scheme_Let_Void;

#define SCHEME_SDFRAME_VAL(obj)    (((Scheme_Static_Dist *)(obj))->frame)
#define SCHEME_SDPOS_VAL(obj)    (((Scheme_Static_Dist *)(obj))->position)

void scheme_add_trigger_callback(Scheme_Trigger_Callback *);
void scheme_add_trigger_from_port(Scheme_Trigger_From_Port *);

#define NOT_BLOCKED 0
#define PIPE_BLOCKED 1
#define PORT_BLOCKED 2
#define TRIGGER_BLOCKED 3

extern int scheme_active_but_sleeping;

typedef struct {
  Scheme_Type type;
  short count;
  Scheme_Object *array[1];
} Scheme_Sequence;


extern Scheme_Object *scheme_quote_symbol;
extern Scheme_Object *scheme_quasiquote_symbol;
extern Scheme_Object *scheme_unquote_symbol;
extern Scheme_Object *scheme_unquote_splicing_symbol;

#ifdef DO_STACK_CHECK
void scheme_init_stack_limit (void);
#endif

void objscheme_setup(Scheme_Env *global_env);

extern Scheme_Object *scheme_not_prim;
extern Scheme_Object *scheme_define_syntax, *scheme_defmacro_syntax;
extern Scheme_Object *scheme_define_struct_syntax;

Scheme_Object *scheme_named_map_1 (char *, Scheme_Object *(*fun)(Scheme_Object*), Scheme_Object *lst);

extern Scheme_Object *scheme_orig_stdout_port;

int scheme_bignum_eq(Scheme_Object *a, Scheme_Object *b);
int scheme_bignum_lt(Scheme_Object *a, Scheme_Object *b);
int scheme_bignum_gt(Scheme_Object *a, Scheme_Object *b);
int scheme_bignum_le(Scheme_Object *a, Scheme_Object *b);
int scheme_bignum_ge(Scheme_Object *a, Scheme_Object *b);
Scheme_Object *scheme_bignum_negate(Scheme_Object *n);
Scheme_Object *scheme_bignum_add(Scheme_Object *a, Scheme_Object *b);
Scheme_Object *scheme_bignum_subtract(Scheme_Object *a, Scheme_Object *b);
Scheme_Object *scheme_bignum_add1(Scheme_Object *n);
Scheme_Object *scheme_bignum_sub1(Scheme_Object *n);
Scheme_Object *scheme_bignum_multiply(Scheme_Object *a, Scheme_Object *b);
Scheme_Object *scheme_bignum_max(Scheme_Object *a, Scheme_Object *b);
Scheme_Object *scheme_bignum_min(Scheme_Object *a, Scheme_Object *b);
void scheme_bignum_divide(Scheme_Object *n, Scheme_Object *d,
			  Scheme_Object **qp, Scheme_Object **rp, int norm);
Scheme_Object *scheme_bignum_power(Scheme_Object *a, Scheme_Object *b);


/* Debugging: */
Scheme_Debugging_Info *scheme_alloc_debug(Scheme_Object *src);
Scheme_Object *scheme_debug_get_last_eval(void);
Scheme_Object *scheme_debug_source(Scheme_Object *debug_handle);

#define scheme_add_good_binding(i,v,f) (f->values[i] = v)

int scheme_equal_structs(Scheme_Object *obj1, Scheme_Object *obj2);

Scheme_Object *scheme_compiled_void(int can_be_value);
Scheme_Object *scheme_standard_evaluator(Scheme_Object *expr);

int scheme_find_type(Scheme_Object *ts);

#define MAX_QUICK_VALUES 12
#if MAX_QUICK_VALUES > 0
typedef struct Scheme_Quick_Env {
  Scheme_Basic_Env basic;
  Scheme_Object *values[MAX_QUICK_VALUES];
} Scheme_Quick_Env;

#define NUM_QUICK_ENVS_PER_SEGMENT 500
typedef struct Scheme_Env_Stack {
  int stack_pos;
  Scheme_Quick_Env array[NUM_QUICK_ENVS_PER_SEGMENT];
  struct Scheme_Env_Stack *next, *prev;
} Scheme_Env_Stack;

Scheme_Env_Stack *scheme_alloc_env_stack(Scheme_Env_Stack *prev);

#define scheme_save_env_stack(ss) \
    (ss.env_stack = scheme_current_process->env_stack, \
     ss.stack_pos = ss.env_stack->stack_pos)

#define scheme_restore_env_stack(ss) \
    (scheme_current_process->env_stack = ss.env_stack, \
     ss.env_stack->stack_pos = ss.stack_pos)

#else
struct Scheme_Env_Stack { int blah; };
scheme_save_env_stack(ss) 0
scheme_restore_env_stack(ss) 0
#endif

#ifndef ERROR_ON_OVERFLOW
extern Scheme_Object *scheme_overflow_reply;
extern Scheme_Jumpup_Buf scheme_overflow_cont;

typedef struct Scheme_Overflow {
  Scheme_Jumpup_Buf cont; /* continuation after value obtained in overflowed */
  struct Scheme_Overflow *prev; /* old overflow info */
  jmp_buf savebuf; /* save old error buffer here */
} Scheme_Overflow;
#endif

void scheme_jmpup_free(Scheme_Jumpup_Buf *);

void scheme_struct_get_arity(Scheme_Object *obj, short *mina, short *maxa);
extern Scheme_Object *scheme_arity_at_least_constr;

extern Scheme_Object *scheme_multiple_values;
extern int scheme_multiple_count;
extern Scheme_Object **scheme_multiple_array;

Scheme_Object *scheme_values(int argc, Scheme_Object *argv[]);
