/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"

#define HASH_TABLE_SIZE 512
static Scheme_Hash_Table *type_symbol_table = NULL;

/* globals */

/* locals */
static Scheme_Object *
tsymbol_p_prim (int argc, Scheme_Object *argv[]);

void
scheme_init_type_symbol_table ()
{
  if (scheme_starting_up)
    type_symbol_table = scheme_hash_table(HASH_TABLE_SIZE, SCHEME_hash_ptr, 
					  0, 1);
}

void
scheme_init_type_symbol (Scheme_Env *env)
{
  scheme_add_global_constant ("type-symbol?", 
			      scheme_make_prim_w_arity(tsymbol_p_prim, 1, 1), 
			      env);
}

Scheme_Object *
scheme_make_type_symbol(Scheme_Object *symbol)
{
  Scheme_Object *sym;

  sym = scheme_alloc_small_object();

  sym->type = scheme_type_symbol_type;

  SCHEME_PTR_VAL(sym) = symbol;
    
  return (sym);
}

Scheme_Object *
scheme_intern_type_symbol (Scheme_Object *symbol)
{
  Scheme_Object *sym;

  sym = (Scheme_Object *)scheme_lookup_in_table(type_symbol_table, 
						(char *)symbol);

  if (!sym) {
    sym = scheme_make_type_symbol(symbol);
    scheme_add_to_table (type_symbol_table, (char *)symbol, sym, 0);
  }

  return (sym);
}

static Scheme_Object *
tsymbol_p_prim (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("type-symbol?", 1, 1, argc, argv);

  return (SCHEME_TSYMBOLP(argv[0]) ? scheme_true : scheme_false);
}

