
; Test/example of wxScheme's object system

(define c1 
  (let ((v 10))
    (make-class '()
		(class (x 1) (y 2))
		(private (a 0) (b3 3))
		(public (b1 2) (b2 2) (e 0))
		(public (c 3) (d 7)
			(f-1-a (lambda () a))
			(f-1-b1 (lambda () b1))
			(f-1-b2 (lambda () b2))
			(f-1-c (lambda () c))
			(f-1-v (lambda () v))
			(f-1-x (lambda () x))
			(f-1-top-a (lambda () (ivar this a)))
			(f-1-other-e (lambda (o) (ivar o e))))
		(lambda (x)
		  (set! e x)
		  (set! a x)))))
  
(define o1 (make-object c1 0))

(define c2 
  (let ((v 20))
    (make-class c1
		(inherit b2)
		(rename (also-b2 b2))
		(public (a 4) (b1 5) (c 6)
			(f-2-a (lambda () a))
			(f-2-b1 (lambda () b1))
			(f-2-b2 (lambda () b2))
			(f-2-also-b2 (lambda () also-b2))
			(f-2-c (lambda () c))
			(f-2-v (lambda () v))
			(f-2-set-b2 (lambda (v) (set! b2 v) b2)))
		(private (y 3))
		(lambda ()
		  (super-init 1))
		(lambda (name)
		  `(unbound ,name)))))

(define o2 (make-object c2))

(define c3
  (make-class* ()
	       (public (x 6) (z 7) (b2 8)
		       (f-3-b2 (lambda () b2)))))

(define c4
  (make-class* ([c1 c1][c3 c3])
	       (inherit c)
	       (rename-from c3 (x z))
	       (public (f-4-c (lambda () c))
		       (f-4-x (lambda () x)))
	       (lambda ()
		 (c1-init 5)
		 (c3-init))))

(define c5
  (make-class* ([c1 c1][c3 c3])
	       (share-from c3 (x z))
	       (share b2)
	       (local (b1 12))
	       (lambda ()
		 (c1-init 5)
		 (c3-init))))

(define o4 (make-object c4))
(define o5 (make-object c5))

(define display-test 
  (lambda (p v)
    (printf "Should be ~s: ~s ~a~n"
	    p v (if (equal? p v)
		    ""
		    "ERROR"))))

(display-test #t (is-a? o1 c1))
(display-test #t (is-a? o2 c1))
(display-test #t (is-a? o4 c1))
(display-test #f (is-a? o1 c2))
(display-test #t (is-a? o2 c2))
(display-test #f (is-a? o4 c2))
(display-test #f (is-a? o1 c3))
(display-test #f (is-a? o2 c3))
(display-test #t (is-a? o4 c3))
(display-test #f (is-a? o1 c4))
(display-test #f (is-a? o2 c4))
(display-test #t (is-a? o4 c4))
(display-test #t (subclass? c2 c1))
(display-test #f (subclass? c1 c2))
(display-test #t (subclass? c4 c1))
(display-test #f (subclass? c1 c4))
(display-test #f (subclass? c1 c3))

(display-test 1 (cvar c1 x))

(define error-1 (lambda () (display-test 'error (ivar o1 a))))
(display-test 4 (ivar o2 a))
(define error-2 (lambda () (display-test 'error (ivar c1 o2 a))))
(display-test 4 (ivar c2 o2 a))

(display-test 0 (send o1 f-1-a))
(display-test 1 (send o2 f-1-a))
(display-test 4 (send o2 f-2-a))
(display-test 5 (send o4 f-1-a))

(define error-3 (lambda () (display-test 'error (send o1 f-1-top-a))))
(display-test 4 (send o2 f-1-top-a))

(display-test 5 (ivar o2 b1))
(display-test '(unbound b3) (ivar o2 b3))
(display-test 2 (ivar c1 o2 b1))
(display-test 5 (ivar c2 o2 b1))

(display-test 2 (send o1 f-1-b1))
(display-test 2 (send o1 f-1-b2))
(display-test 5 (send o2 f-1-b1))
(display-test 2 (send o2 f-1-b2))
(display-test 5 (send o2 f-2-b1))
(display-test 2 (send o2 f-2-b2))
(display-test 2 (send o2 f-2-also-b2))

(display-test 3 (ivar o1 c))
(display-test 6 (ivar o2 c))
(display-test 3 (ivar c1 o2 c))
(display-test 6 (ivar c2 o2 c))

(display-test 3 (send o1 f-1-c))
(display-test 6 (send o2 f-1-c))
(display-test 6 (send o2 f-2-c))
(display-test 3 (send o4 f-4-c))

(display-test 7 (ivar o1 d))
(display-test 7 (ivar o2 d))

(display-test 10 (send o1 f-1-v))
(display-test 10 (send o2 f-1-v))
(display-test 20 (send o2 f-2-v))

(display-test 0 (send o1 f-1-other-e o1))
(display-test 1 (send o1 f-1-other-e o2))

(display-test 2 (ivar o2 y))

(display-test 3 (send o2 f-2-set-b2 3))
(display-test 3 (send o2 f-2-also-b2))

(display-test 1 (ivar o4 x))
(display-test 7 (send o4 f-4-x))
(display-test 1 (ivar c1 o4 x))
(display-test 6 (ivar c3 o4 x))

(display-test 7 (ivar o5 x))
(display-test 7 (send o5 f-1-x))
(display-test 12 (ivar o5 b1))
(display-test 2 (send o5 f-1-b1))
(display-test 2 (send o5 f-1-b2))
(display-test 2 (send o5 f-3-b2))

(define error-4 (lambda () (display-test 'error (ivar c1 o2 A))))

(define g1 (make-generic c1 x))
(display-test 1 (g1 o1))
(display-test 1 (g1 o2))
(display-test 1 (g1 o4))
(display-test 7 (g1 o5))

(define g2 (make-generic c2 x))
(display-test 1 (g2 o2))

(define error-5 (lambda () (display-test 'error (g2 o1))))

(define c100
  (let loop ([n 99][c c1])
    (if (zero? n)
	c
	(loop (sub1 n) (make-class c (public (z n)))))))

(define c1024
  (let loop ([n 9][c c1])
    (if (zero? n)
	c
	(loop (sub1 n) (make-class* ([first c][second c])
				    (rename-from first [ff f-1-b1])
				    (rename-from second [fs f-1-b1])
				    (public
				     [f-1-b1
				      (lambda ()
					(+ (ff) (fs)))]))))))

(define test-error 
  (lambda (thunk)
    (let ((old-error-handler (error-handler)))
      (call/ec
       (lambda (exit)
	 (let ((new-error-handler (lambda errdata
				    (exit #f))))
	   (error-handler new-error-handler)
	   (thunk))))
      (error-handler old-error-handler))))

; Nothing should be printed from this:
(map test-error (list error-1 error-2 error-3 error-4 error-5))
